
from ..pitch import NotePitch


C_SHARP_0 = NotePitch('C', '#' , 0)
D_SHARP_0 = NotePitch('D', '#' , 0)
E_SHARP_0 = NotePitch('E', '#' , 0)
F_SHARP_0 = NotePitch('F', '#' , 0)
G_SHARP_0 = NotePitch('G', '#' , 0)
A_SHARP_0 = NotePitch('A', '#' , 0)
B_SHARP_0 = NotePitch('B', '#' , 0)

C_SHARP_1 = NotePitch('C', '#' , 1)
D_SHARP_1 = NotePitch('D', '#' , 1)
E_SHARP_1 = NotePitch('E', '#' , 1)
F_SHARP_1 = NotePitch('F', '#' , 1)
G_SHARP_1 = NotePitch('G', '#' , 1)
A_SHARP_1 = NotePitch('A', '#' , 1)
B_SHARP_1 = NotePitch('B', '#' , 1)

C_SHARP_2 = NotePitch('C', '#' , 2)
D_SHARP_2 = NotePitch('D', '#' , 2)
E_SHARP_2 = NotePitch('E', '#' , 2)
F_SHARP_2 = NotePitch('F', '#' , 2)
G_SHARP_2 = NotePitch('G', '#' , 2)
A_SHARP_2 = NotePitch('A', '#' , 2)
B_SHARP_2 = NotePitch('B', '#' , 2)

C_SHARP_3 = NotePitch('C', '#' , 3)
D_SHARP_3 = NotePitch('D', '#' , 3)
E_SHARP_3 = NotePitch('E', '#' , 3)
F_SHARP_3 = NotePitch('F', '#' , 3)
G_SHARP_3 = NotePitch('G', '#' , 3)
A_SHARP_3 = NotePitch('A', '#' , 3)
B_SHARP_3 = NotePitch('B', '#' , 3)

C_SHARP_4 = NotePitch('C', '#' , 4)
D_SHARP_4 = NotePitch('D', '#' , 4)
E_SHARP_4 = NotePitch('E', '#' , 4)
F_SHARP_4 = NotePitch('F', '#' , 4)
G_SHARP_4 = NotePitch('G', '#' , 4)
A_SHARP_4 = NotePitch('A', '#' , 4)
B_SHARP_4 = NotePitch('B', '#' , 4)

C_SHARP_5 = NotePitch('C', '#' , 5)
D_SHARP_5 = NotePitch('D', '#' , 5)
E_SHARP_5 = NotePitch('E', '#' , 5)
F_SHARP_5 = NotePitch('F', '#' , 5)
G_SHARP_5 = NotePitch('G', '#' , 5)
A_SHARP_5 = NotePitch('A', '#' , 5)
B_SHARP_5 = NotePitch('B', '#' , 5)

C_SHARP_6 = NotePitch('C', '#' , 6)
D_SHARP_6 = NotePitch('D', '#' , 6)
E_SHARP_6 = NotePitch('E', '#' , 6)
F_SHARP_6 = NotePitch('F', '#' , 6)
G_SHARP_6 = NotePitch('G', '#' , 6)
A_SHARP_6 = NotePitch('A', '#' , 6)
B_SHARP_6 = NotePitch('B', '#' , 6)

C_SHARP_7 = NotePitch('C', '#' , 7)
D_SHARP_7 = NotePitch('D', '#' , 7)
E_SHARP_7 = NotePitch('E', '#' , 7)
F_SHARP_7 = NotePitch('F', '#' , 7)
G_SHARP_7 = NotePitch('G', '#' , 7)
A_SHARP_7 = NotePitch('A', '#' , 7)
B_SHARP_7 = NotePitch('B', '#' , 7)

C_SHARP_8 = NotePitch('C', '#' , 8)
D_SHARP_8 = NotePitch('D', '#' , 8)
E_SHARP_8 = NotePitch('E', '#' , 8)
F_SHARP_8 = NotePitch('F', '#' , 8)
G_SHARP_8 = NotePitch('G', '#' , 8)
A_SHARP_8 = NotePitch('A', '#' , 8)
B_SHARP_8 = NotePitch('B', '#' , 8)

C_SHARP_9 = NotePitch('C', '#' , 9)
D_SHARP_9 = NotePitch('D', '#' , 9)
E_SHARP_9 = NotePitch('E', '#' , 9)
F_SHARP_9 = NotePitch('F', '#' , 9)
G_SHARP_9 = NotePitch('G', '#' , 9)
A_SHARP_9 = NotePitch('A', '#' , 9)
B_SHARP_9 = NotePitch('B', '#' , 9)
